/*
 * MZViewer - A viewer for MZData format Mass Spectra
 * Copyright (c) 2005 Simon Andrews

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */
package uk.ac.bbsrc.babraham.mzdata.parser;
import net.sourceforge.iharder.Base64;

import org.xml.sax.*;
import org.xml.sax.helpers.*;
import uk.ac.bbsrc.babraham.mzdata.*;
import uk.ac.bbsrc.babraham.mzdata.displays.*;
import uk.ac.bbsrc.babraham.mzdata.exceptions.*;

import java.io.*;
import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;


/**
 * @author andrewss
 *
 */
public class MZDataReader implements Runnable {

	private Vector listeners;
	private InputSource input;
	private int currentSpectrumCount = 0;
	private MZCollection collection;
	private Exception exception = null;
	    
    public MZDataReader () {
		listeners = new Vector();
    }
        
    public MZCollection readMZFile (String filename) throws IOException, SAXException {
            currentSpectrumCount = 0;
    		Thread t = new Thread(this);
    		collection = new MZCollection();
    		input = new InputSource(new FileInputStream(new File(filename)));
    		input.setSystemId(filename);
    		t.start();
    		return collection;
    }
    
    public MZCollection readMZFile (URL url) throws IOException, SAXException {
        currentSpectrumCount = 0;
		Thread t = new Thread(this);
		collection = new MZCollection();
		input = new InputSource(url.openStream());
		input.setSystemId(url.toString());
		t.start();
		return collection;
    }
    
    
    public void run () {
    		try {	
			XMLReader reader = XMLReaderFactory.createXMLReader("org.apache.xerces.parsers.SAXParser");

			MZContentHandler handler = new MZContentHandler();

			reader.setContentHandler(handler);
			reader.parse(input);
    		}
    		catch (Exception ex) {
//    			ex.printStackTrace();
    			Enumeration e = listeners.elements();
    			while (e.hasMoreElements()) {
    				((LoadProgressListener)e.nextElement()).loadError(ex);
    			}

    		}
		
		Enumeration e = listeners.elements();
		while (e.hasMoreElements()) {
			((LoadProgressListener)e.nextElement()).loadComplete();
		}
            	
    }

    
    private class MZContentHandler implements ContentHandler {
        
        private boolean inSpectrum = false;
        private boolean inData = false;
        private StringBuffer dataBuffer;
        private boolean isBigEndian;
        private int dataPrecision;
        private int expectedLength;
        private MZSpectrum spectrum;
                
        public void setDocumentLocator(Locator locator) {
        }

        public void startDocument() throws SAXException {
        }

        public void endDocument() throws SAXException {
        }

        public void startPrefixMapping(String arg0, String arg1) throws SAXException {
  
        }

        public void endPrefixMapping(String arg0) throws SAXException {    
        }

        public void startElement(String namespaceURI, String localName, String qName, Attributes attribs) throws SAXException {
 
            if (localName.equals("spectrum")) {
                spectrum = new MZSpectrum();
                spectrum.setID(Integer.parseInt(attribs.getValue("id")));
                inSpectrum = true;
            }
            else if (localName.equals("spectrumList")) {
                int count = Integer.parseInt(attribs.getValue("count"));
                try {
                    collection.setSpectrumCount(count);
                }
                catch (MZCollectionException e) {
                    throw new SAXException(e.getMessage());
                }
            }
            if (inSpectrum) {
                if (localName.equals("data")) {
                    dataPrecision = Integer.parseInt(attribs.getValue("precision"));
                    if (attribs.getValue("endian").equals("big")) {
                        isBigEndian = true;
                    }
                    else {
                        isBigEndian = false;
                    }
                    expectedLength = Integer.parseInt(attribs.getValue("length"));
                    dataBuffer = new StringBuffer();
                    inData = true;                    
                }
                else if (localName.equals("acqInstrument")  || localName.equals("spectrumInstrument")) {
                		spectrum.setMsLevel(Integer.parseInt(attribs.getValue("msLevel")));
                }
                else if (localName.equals("acqSpecification")) {
                		try {
                			spectrum.setType(attribs.getValue("spectrumType"));
                		}
                		catch (MZSpectrumException e) {
                			throw new SAXException(e.getMessage());
                		}
                }
                else if (localName.equals("precursor")) {
                    int spectrumRef = Integer.parseInt(attribs.getValue("spectrumRef"));
                    MZSpectrum parent = collection.getSpectrumWithID(spectrumRef);
                    if (parent != null) {
                        parent.addChild(spectrum);
                        spectrum.addParent(parent);
                    }
                }
                else if (localName.equals("cvParam")){
                 	String name = attribs.getValue("name");
                 	// This is for mzData 1.05
                 	if (name != null && name.startsWith("TimeIn")){
                 		String unit = "minutes";
                 		if (name.equals("TimeInSeconds")) {
                 			unit = "seconds";
                 		}
                 		try {
                 			spectrum.setTime(Float.parseFloat(attribs.getValue("value")),unit);
                 		}
                 		catch (MZSpectrumException e) {
                 			throw new SAXException (e.getMessage());
                 		}
                 	}
                 	// This is for mzData 1.04
                 	else if (name != null && name.startsWith("time.")) {
                 	    String unit = "seconds";
                 	    if (name.equals("time.min")) {
                 	        unit = "minutes";
                 	    }
                 		try {
                 			spectrum.setTime(Float.parseFloat(attribs.getValue("value")),unit);
                 		}
                 		catch (MZSpectrumException e) {
                 			throw new SAXException (e.getMessage());
                 		}
                 	}
                    
                    // This retrieves the parent mz ratio of any MS2 spectra
                    else if (name != null && name.equals("MassToChargeRatio")) {
                        spectrum.setParentMZ(Float.parseFloat(attribs.getValue("value")));
                    }
                } 
            }
        }

        public void endElement(String namespaceURI, String localName, String qName) throws SAXException {
            if (localName.equals("spectrum")) {
                try {
                    collection.addSpectrum(spectrum);
                    ++currentSpectrumCount;
                    if (currentSpectrumCount % 20 == 0) {
	                    Enumeration e = listeners.elements();
	                    while (e.hasMoreElements()) {
	                    	((LoadProgressListener)e.nextElement()).updateLoadCount(currentSpectrumCount,collection.getSpectrumCount());
	                    }
	                    try {
							Thread.sleep(100);
						} catch (InterruptedException e1) {
							e1.printStackTrace();
						}
                    }
                }
                catch (MZCollectionException e) {
                    throw new SAXException(e.getMessage());
                }
                inSpectrum = false;
            }
            else if (localName.equals("data")) {
                byte [] dataBytes = Base64.decode(dataBuffer.toString());
                                
                // We need to convert this from a byte array to a float.
                // The method we use depends on whether the input data is
                // big or little endian and whether the float is encoded
                // in 32 or 64 bits

                if (dataPrecision == 32) {
                       
	                if (dataBytes.length%4 != 0) {
	                    throw new SAXException("Byte data array had length "+dataBytes.length+" which isn't a multiple of 4");
	                }
	         
	                float [] dataFloats= new float[dataBytes.length/4];
	                	                
	                if (dataFloats.length != expectedLength) {
	                    if (expectedLength != 0) {
	                        // Empty spectra are denoted by an expected length of 0 and a single data point of 0:0
	                        throw new SAXException ("Expected data length of "+expectedLength+" but got actual length of "+dataFloats.length);
	                    }
	                }
	                
	               
	                for (int i=0;i<=(dataBytes.length-4);i+=4) {
	                   int myInt; 
	                   if (isBigEndian) {
	                       myInt = (dataBytes[i] << 24) |
	                       ((dataBytes[i+1] & 0xff) << 16) |
	                       ((dataBytes[i+2] & 0xff) << 8) |
	                       (dataBytes[i+3] & 0xff);
	                   }
	                   else {
	                       myInt = (dataBytes[i+3] << 24) |
	                       ((dataBytes[i+2] & 0xff) << 16) |
	                       ((dataBytes[i+1] & 0xff) << 8) |
	                       (dataBytes[i] & 0xff);
	                   }
	                   
	                   dataFloats[i/4] = Float.intBitsToFloat(myInt);
	                }

	                try {
	                    spectrum.addFloatArray(dataFloats);
	                }
	                catch (MZSpectrumException e) {
	                    throw new SAXException (e.getMessage());
	                }
	            }
                
                else if (dataPrecision == 64) {
                    
	                if (dataBytes.length%8 != 0) {
	                    throw new SAXException("Byte data array had length "+dataBytes.length+" which isn't a multiple of 8");
	                }
	         
	                double [] dataDoubles= new double[dataBytes.length/8];
	                	                
	                if (dataDoubles.length != expectedLength) {
	                    if (expectedLength != 0) {
	                        // Empty spectra are denoted by an expected length of 0 and a single data point of 0:0
	                        throw new SAXException ("Expected data length of "+expectedLength+" but got actual length of "+dataDoubles.length);
	                    }
	                }
	               
	                for (int i=0;i<=(dataBytes.length-8);i+=8) {
                       long myLong; 
	                   if (isBigEndian) {
	                       myLong = (((long)dataBytes[i]) << 56) |
	                       (((long)(dataBytes[i+1] & 0xff)) << 48) |
	                       (((long)(dataBytes[i+2] & 0xff)) << 40) |
	                       (((long)(dataBytes[i+3] & 0xff)) << 32) |	                               
	                       (((long)(dataBytes[i+4] & 0xff)) << 24) |
	                       (((long)(dataBytes[i+5] & 0xff)) << 16) |
	                       (((long)(dataBytes[i+6] & 0xff)) << 8) |
                           ((long)(dataBytes[i+7] & 0xff));
                       }
	                   else {
	                       myLong = ((long)(dataBytes[i+7]) << 56) |
	                       (((long)(dataBytes[i+6] & 0xff)) << 48) |
                           (((long)(dataBytes[i+5] & 0xff) << 40)) |
                           (((long)(dataBytes[i+4] & 0xff) << 32)) |	                               
                           (((long)(dataBytes[i+3] & 0xff) << 24)) |
                           (((long)(dataBytes[i+2] & 0xff) << 16)) |
                           (((long)(dataBytes[i+1] & 0xff) << 8)) |
                           ((long)(dataBytes[i] & 0xff));
	                   }
	                   dataDoubles[i/8] = Double.longBitsToDouble(myLong);
	                }
	                try {
	                    spectrum.addDoubleArray(dataDoubles);
	                }
	                catch (MZSpectrumException e) {
	                    throw new SAXException (e.getMessage());
	                }
                }
                
                inData = false;

            } 
            else if (localName.equals("spectrumList")) {
            		try {
            			collection.checkSpecturmCount();
            		}
            		catch (MZCollectionException e) {
            			throw new SAXException(e.getMessage());
            		}
            }
        }

        public void characters(char[] chars, int start, int length) throws SAXException {
            if (inData) {
                
                // Strip out any newlines or whitespace.  This isn't strictly
                // necessary if using the iharder Base64 library as it handles
                // this, but other libraries might not be so forgiving.
                char [] in = (new String(chars,start,length)).toCharArray();
                StringBuffer out = new StringBuffer();
                
                for (int i=0;i<in.length;i++) {
                    switch (in[i]) {
                        case '\r':
                        case '\n':
                        case ' ':
                        case '\t':
                        // Throw away
                        break;
                    default:
                        out.append(in[i]);
                    }
                }
                dataBuffer.append(out.toString());
            }            
        }

        public void ignorableWhitespace(char[] arg0, int arg1, int arg2) throws SAXException {
        }

        public void processingInstruction(String arg0, String arg1) throws SAXException {
        }

        public void skippedEntity(String arg0) throws SAXException {
        }
        
    }

    public void addLoadProgressListener (LoadProgressListener l) {
    		listeners.add(l);
    }

    	public void removeLoadProgressListener (LoadProgressListener l) {
    		listeners.remove(l);
    	}
    
}
